@@ -17,7 +17,7 @@ import unittest
 import gnupg
 
 __author__ = "Vinay Sajip"
-__date__  = "$05-Feb-2014 17:48:48$"
+__date__  = "$07-Dec-2014 18:46:40$"
 
 ALL_TESTS = True
 
@@ -106,6 +106,11 @@ class GPGTestCase(unittest.TestCase):
                 gpg.options = ['--debug-quick-random']
             else:
                 gpg.options = ['--quick-random']
+        self.test_fn = test_fn = 'random_binary_data'
+        if not os.path.exists(test_fn):
+            data_file = open(test_fn, 'wb')
+            data_file.write(os.urandom(5120 * 1024))
+            data_file.close()
 
     def test_environment(self):
         "Test the environment by ensuring that setup worked"
@@ -147,7 +152,7 @@ class GPGTestCase(unittest.TestCase):
         params['Passphrase'] = passphrase
         cmd = self.gpg.gen_key_input(**params)
         return self.gpg.gen_key(cmd)
-    
+
     def do_key_generation(self):
         "Test that key generation succeeds"
         result = self.generate_key("Barbara", "Brown", "beta.com")
@@ -217,7 +222,7 @@ class GPGTestCase(unittest.TestCase):
         params['key_type'] = 'DSA'
         cmd = self.gpg.gen_key_input(**params)
         self.assertTrue('Key-Type: DSA\n' in cmd)
-        
+
     def test_list_keys_after_generation(self):
         "Test that after key generation, the generated key is available"
         self.test_list_keys_initial()
@@ -238,6 +243,20 @@ class GPGTestCase(unittest.TestCase):
         private_keys_2 = gpg.list_keys(secret=True)
         self.assertEqual(private_keys_2, private_keys)
 
+    def test_scan_keys(self):
+        "Test that external key files can be scanned"
+        expected = set([
+            'Andrew Able (A test user) <andrew.able@alpha.com>',
+            'Barbara Brown (A test user) <barbara.brown@beta.com>',
+            'Charlie Clark (A test user) <charlie.clark@gamma.com>',
+        ])
+        for fn in ('test_pubring.gpg', 'test_secring.gpg'):
+            data = self.gpg.scan_keys(fn)
+            uids = set()
+            for d in data:
+                uids.add(d['uids'][0])
+            self.assertEqual(uids, expected)
+
     def test_encryption_and_decryption(self):
         "Test that encryption and decryption works"
         logger.debug("test_encryption_and_decryption begins")
@@ -273,11 +292,12 @@ class GPGTestCase(unittest.TestCase):
         ddata = gpg.decrypt(edata, passphrase='bbrown')
         self.assertEqual(data, str(ddata))
         # Test symmetric encryption with non-default cipher
-        data = "chippy was here"
         edata = str(gpg.encrypt(data, None, passphrase='bbrown',
                     symmetric='AES256'))
         ddata = gpg.decrypt(edata, passphrase='bbrown')
         self.assertEqual(data, str(ddata))
+        # Test that you can't encrypt with no recipients
+        self.assertRaises(ValueError, self.gpg.encrypt, data, [])
 
     def test_import_and_export(self):
         "Test that key import and export works"
@@ -353,12 +373,7 @@ class GPGTestCase(unittest.TestCase):
                          "Fingerprints must match")
         self.assertEqual(verified.trust_level, verified.TRUST_ULTIMATE)
         self.assertEqual(verified.trust_text, 'TRUST_ULTIMATE')
-        test_fn = 'random_binary_data'
-        if not os.path.exists(test_fn):
-            data_file = open(test_fn, 'wb')
-            data_file.write(os.urandom(5120 * 1024))
-            data_file.close()
-        data_file = open(test_fn, 'rb')
+        data_file = open(self.test_fn, 'rb')
         sig = self.gpg.sign_file(data_file, keyid=key.fingerprint,
                                  passphrase='aable')
         data_file.close()
@@ -375,7 +390,7 @@ class GPGTestCase(unittest.TestCase):
             logger.debug("ver: %r", verified.fingerprint)
         self.assertEqual(key.fingerprint, verified.fingerprint,
                          "Fingerprints must match")
-        data_file = open(test_fn, 'rb')
+        data_file = open(self.test_fn, 'rb')
         sig = self.gpg.sign_file(data_file, keyid=key.fingerprint,
                                  passphrase='aable', detach=True)
         data_file.close()
@@ -383,7 +398,7 @@ class GPGTestCase(unittest.TestCase):
         self.assertTrue(sig.hash_algo)
         try:
             file = gnupg._make_binary_stream(sig.data, self.gpg.encoding)
-            verified = self.gpg.verify_file(file, test_fn)
+            verified = self.gpg.verify_file(file, self.test_fn)
         except UnicodeDecodeError: #happens in Python 2.6
             from io import BytesIO
             verified = self.gpg.verify_file(BytesIO(sig.data))
@@ -393,7 +408,7 @@ class GPGTestCase(unittest.TestCase):
         self.assertEqual(key.fingerprint, verified.fingerprint,
                          "Fingerprints must match")
         # Test in-memory verification
-        data_file = open(test_fn, 'rb')
+        data_file = open(self.test_fn, 'rb')
         data = data_file.read()
         data_file.close()
         fd, fn = tempfile.mkstemp()
@@ -410,6 +425,34 @@ class GPGTestCase(unittest.TestCase):
                          "Fingerprints must match")
         logger.debug("test_signature_verification ends")
 
+    def test_signature_file(self):
+        "Test that signing and verification works via the GPG output"
+        logger.debug("test_signature_file begins")
+        key = self.generate_key("Andrew", "Able", "alpha.com")
+        data_file = open(self.test_fn, 'rb')
+        sig_file = self.test_fn + '.asc'
+        sig = self.gpg.sign_file(data_file, keyid=key.fingerprint,
+                                 passphrase='aable', detach=True,
+                                 output=sig_file)
+        data_file.close()
+        self.assertTrue(sig, "File signing should succeed")
+        self.assertTrue(sig.hash_algo)
+        self.assertTrue(os.path.exists(sig_file))
+        # Test in-memory verification
+        data_file = open(self.test_fn, 'rb')
+        data = data_file.read()
+        data_file.close()
+        try:
+            verified = self.gpg.verify_data(sig_file, data)
+        finally:
+            os.unlink(sig_file)
+        if key.fingerprint != verified.fingerprint:
+            logger.debug("key: %r", key.fingerprint)
+            logger.debug("ver: %r", verified.fingerprint)
+        self.assertEqual(key.fingerprint, verified.fingerprint,
+                         "Fingerprints must match")
+        logger.debug("test_signature_file ends")
+
     def test_deletion(self):
         "Test that key deletion works"
         logger.debug("test_deletion begins")
@@ -425,14 +468,7 @@ class GPGTestCase(unittest.TestCase):
 
     def test_nogpg(self):
         "Test that absence of gpg is handled correctly"
-        # We use shell=False on POSIX, shell=True on Windows
-        # and this can lead to one of two possible exceptions
-        # depending on the platform.
-        if os.name == 'posix':
-            exc = OSError
-        else:
-            exc = ValueError
-        self.assertRaises(exc, gnupg.GPG, gnupghome=self.homedir,
+        self.assertRaises(OSError, gnupg.GPG, gnupghome=self.homedir,
                           gpgbinary='frob')
 
     def test_make_args(self):
@@ -442,20 +478,16 @@ class GPGTestCase(unittest.TestCase):
         self.assertTrue(len(args) > 4)
         self.assertEqual(args[-4:], ['--foo', '--bar', 'a', 'b'])
 
-    def test_file_encryption_and_decryption(self):
-        "Test that encryption/decryption to/from file works"
-        logger.debug("test_file_encryption_and_decryption begins")
-        encfno, encfname = tempfile.mkstemp()
-        decfno, decfname = tempfile.mkstemp()
+    def do_file_encryption_and_decryption(self, encfname, decfname):
+        "Do the actual encryption.decryptin test using given filenames"
+        mode = None
         if os.name == 'posix':
             # pick a mode that won't be already in effect via umask
-            mode = os.stat(encfname).st_mode | stat.S_IXUSR
-            os.chmod(encfname, mode)
-            # assume same for decfname
-            os.chmod(decfname, mode)
-        # On Windows, if the handles aren't closed, the files can't be deleted
-        os.close(encfno)
-        os.close(decfno)
+            if os.path.exists(encfname) and os.path.exists(decfname):
+                mode = os.stat(encfname).st_mode | stat.S_IXUSR
+                os.chmod(encfname, mode)
+                # assume same for decfname
+                os.chmod(decfname, mode)
         logger.debug('Encrypting to: %r', encfname)
         logger.debug('Decrypting to: %r', decfname)
         try:
@@ -485,14 +517,36 @@ class GPGTestCase(unittest.TestCase):
             self.assertEqual(data, ddata, "Round-trip must work")
         finally:
             for fn in (encfname, decfname):
-                if os.name == 'posix':
+                if os.name == 'posix' and mode is not None:
                     # Check that the file wasn't deleted, and that the
                     # mode bits we set are still in effect
                     self.assertEqual(os.stat(fn).st_mode, mode)
                 if os.path.exists(fn):
                     os.remove(fn)
+
+    def test_file_encryption_and_decryption(self):
+        "Test that encryption/decryption to/from file works"
+        logger.debug("test_file_encryption_and_decryption begins")
+        encfno, encfname = tempfile.mkstemp()
+        decfno, decfname = tempfile.mkstemp()
+        # On Windows, if the handles aren't closed, the files can't be deleted
+        os.close(encfno)
+        os.close(decfno)
+        self.do_file_encryption_and_decryption(encfname, decfname)
         logger.debug("test_file_encryption_and_decryption ends")
 
+    def test_filenames_with_spaces(self):       # See Issue #16
+        "Test that filenames with spaces are correctly handled"
+        logger.debug("test_filename_with_spaces begins")
+        d = tempfile.mkdtemp()
+        try:
+            encfname = os.path.join(d, 'encrypted file')
+            decfname = os.path.join(d, 'decrypted file')
+            self.do_file_encryption_and_decryption(encfname, decfname)
+        finally:
+            shutil.rmtree(d)
+        logger.debug("test_filename_with_spaces ends")
+
     def test_search_keys(self):
         "Test that searching for keys works"
         r = self.gpg.search_keys('<vinay_sajip@hotmail.com>')
@@ -527,29 +581,46 @@ class GPGTestCase(unittest.TestCase):
         finally:
             shutil.rmtree(workdir)
 
+    def disabled_test_signing_with_uid(self):
+        "Test that signing with uids works. On hold for now."
+        logger.debug("test_signing_with_uid begins")
+        key = self.generate_key("Andrew", "Able", "alpha.com")
+        uid = self.gpg.list_keys(True)[-1]['uids'][0]
+        try:
+            signfile = open(self.test_fn,'rb')
+            signed = self.gpg.sign_file(signfile, keyid=uid,
+                                        passphrase='aable',
+                                        detach=True)
+        finally:
+            signfile.close()
+        self.assertTrue(signed.data)
+        logger.debug("test_signing_with_uid ends")
+
 TEST_GROUPS = {
     'sign' : set(['test_signature_verification']),
     'crypt' : set(['test_encryption_and_decryption',
-                   'test_file_encryption_and_decryption']),
+                   'test_file_encryption_and_decryption',
+                   'test_filenames_with_spaces']),
     'key' : set(['test_deletion', 'test_import_and_export',
                  'test_list_keys_after_generation',
                  'test_key_generation_with_invalid_key_type',
                  'test_key_generation_with_escapes',
                  'test_key_generation_with_empty_value',
                  'test_key_generation_with_colons',
-                 'test_search_keys']),
+                 'test_search_keys', 'test_scan_keys']),
     'import' : set(['test_import_only']),
     'basic' : set(['test_environment', 'test_list_keys_initial',
                    'test_nogpg', 'test_make_args',
                    'test_quote_with_shell']),
+    'test': set(['test_search_keys']),
 }
 
 def suite(args=None):
     if args is None:
         args = sys.argv[1:]
-    if not args:
+    if not args or args == ['--no-doctests']:
         result = unittest.TestLoader().loadTestsFromTestCase(GPGTestCase)
-        want_doctests = True
+        want_doctests = not args
     else:
         tests = set()
         want_doctests = False
@@ -559,7 +630,7 @@ def suite(args=None):
             elif arg == "doc":
                 want_doctests = True
             else:
-                print("Ignoring unknown test group %r" % arg)        
+                print("Ignoring unknown test group %r" % arg)
         result = unittest.TestSuite(list(map(GPGTestCase, tests)))
     if want_doctests:
         result.addTest(doctest.DocTestSuite(gnupg))
